

import java.util.*;

/**
 * This is a suite of benchmarks that are relatively short, both
 * in program size and execution time.  It requires no input, and
 * prints the execution time for each program.  It does a
 * rudimentary check to make sure each program gets the right
 * output.<p>
 *
 * These programs were gathered by John Hennessy and modified by
 * Peter Nye and ported to Java by Jon Maxwell.  This version
 * optimized for Java but without changing the algorithm.<p>
 */

// GVG (Greg Gritton): Will need to increase ITER [JAM: above 500?]


public final class BenchOpt {

  public static final int ITER_MULT = 10;

  public static final double bubblebase = 1.61;
  public static final double dnfbase = 3.5;
  public static final double permbase = 1.75;
  public static final double queensbase = 1.83;
  public static final double towersbase = 2.39;
  public static final double quickbase = 1.92;
  public static final double intmmbase = 1.46;
  public static final double treebase = 2.5;
  public static final double mmbase = 0.0 /* 0.73 */;
  public static final double fpmmbase = 2.92;
  public static final double puzzlebase = 0.5;
  public static final double fftbase = 0.0 /* 1.11 */;
  public static final double fpfftbase = 4.44;

  static double value;
  static double fixed, floated, total;
  static double st_total; /* Total time for same # of iterations as Smalltalk runs */

  static String format(double value, int places, int decimals) {
    return ""+value; // fix to use formatter
  }

  static void runTest(String name, Benchmark benchmark, double fixed_base, 
                      double float_base, int iterations) {
    long starttime;  /* starting time for test */
    double runtime;  /* total time the test runs */
    double avgtime;  /* time per run */

    // System.gc();    - GVG - if the test generates garbage, the GC time needs to be counted
    starttime = System.currentTimeMillis();
    for (int i=0; i<iterations; i++) {
      benchmark.run();
    }
    runtime = (System.currentTimeMillis()-starttime);
    avgtime = runtime / iterations;
    // System.gc();

    System.out.println(name+":  " + format(iterations,4,0) + " iterations at "
                       + format(avgtime,6,2)+" each.");
    fixed += fixed_base * avgtime;
    floated += float_base * avgtime;
    total += runtime;
  }

  public static void main(String args[]) {
    fixed = floated = total = 0.0;

    System.out.println("Reporting timer per iteration in milliseconds.\n");

    /* output results */
    runTest("  Bubble", new BubbleSort(),           bubblebase, bubblebase, 10 * ITER_MULT);
    runTest("   Intmm", new IntMatrixMultiply(),    intmmbase , intmmbase , 10 * ITER_MULT);
    runTest("      Mm", new DoubleMatrixMultiply(), mmbase    , fpmmbase  , 10 * ITER_MULT);
    runTest("    Perm", new Perm(),                 permbase  , permbase  , 10 * ITER_MULT);
    runTest("  Queens", new Queen(),                queensbase, queensbase, 10 * ITER_MULT);
    runTest("   Quick", new QuickSort(),            quickbase , quickbase , 10 * ITER_MULT);
    runTest("    Tree", new TreeSort(),             treebase  , treebase  , 10 * ITER_MULT);
    runTest("  Towers", new Towers(),               towersbase, towersbase, 10 * ITER_MULT);
    runTest("     FFT", new FFT(),                  fftbase   , fpfftbase ,  5 * ITER_MULT);
    runTest("  Puzzle", new Puzzle(),               puzzlebase, puzzlebase,  3 * ITER_MULT);

    /* compute composites */
    System.out.println("");
    System.out.println("Total                            " + format(total, 5, 0));
    System.out.println("Equivalent Total                 " + format(total / ITER_MULT, 5, 0));
    System.out.println("Nonfloating point composite is   " + format(fixed, 7, 2));
    System.out.println("Floating point composite is      " + format(floated, 7, 2));
    System.out.println("");
  }

}


abstract class Benchmark {
  public static final int falseInt = 0;
  public static final int trueInt = 1;
  static int seed;

  static void initRandom() {
    seed = 74755;
  };

  static int randomInt() {
    seed = (seed * 1309 + 13849) & 65535;
    return seed;
  };

  abstract public void run();
};










final class Towers extends Benchmark {
  public static final int maxcells = 18;
  public static final int stackrange = 3;

  /* discsizrange = 1..maxcells; */
  /*    cellcursor = 0..maxcells; */

  static final class Element {
    int discsize;
    int next;
  };

  int freelist, movesdone;
  int stack[] = new int[stackrange+1];
  Element cellspace[] = new Element[maxcells+1];

  void error(String emsg) {
    System.out.println(" Error in Towers: "+emsg);
  };

  void makenull(int s) {
    stack[s]=0;
  };

  int getElement() {
    int temp = 0;
    if ( freelist>0 ) {
      temp = freelist;
      freelist = cellspace[freelist].next;
    }
    else
      error("out of space   ");
    return temp;
  };

  void push(int i,int s) {
    boolean errorfound = false;
    int stackS = stack[s];

    if ( stackS > 0 )
      if ( cellspace[stackS].discsize<=i ) {
        errorfound=true;
        error("disc size error");
      };
    if ( !errorfound ) {
      int localel=getElement();
      Element e = cellspace[localel];
      e.next=stackS;
      e.discsize=i;
      stack[s]=localel;
    }
  };

  void init(int s, int n) {
    int discctr;
    makenull(s);
    for ( discctr = n; discctr >= 1; discctr-- )
      push(discctr,s);
  };

  int pop(int s) {
    int temp, temp1;
    int stackS = stack[s];

    if ( stackS > 0 ) {
      Element e = cellspace[stackS];
      temp1 = e.discsize;
      temp = e.next;
      e.next=freelist;
      freelist=stackS;
      stack[s]=temp;
      return temp1;
    }
    else
      error("nothing to pop ");
    return 0;
  };

  void move(int s1, int s2) {
    push(pop(s1),s2);
    movesdone=movesdone+1;
  };

  void tower(int i, int j, int k) {
    int other;
    if ( k==1 )
      move(i,j);
    else {
      other=6-i-j;
      tower(i,other,k-1);
      move(i,j);
      tower(other,j,k-1);
    }
  };

  public void run() {
    for (int i=1; i <= maxcells; i++) {
      cellspace[i] = new Element();
      cellspace[i].next=i-1;
    }
    freelist=maxcells;
    init(1,14);
    makenull(2);
    makenull(3);
    movesdone=0;
    tower(1,2,14);
    if ( movesdone != 16383 )
      System.out.println (" Error in Towers.");
  };
}

final class IntMatrixMultiply extends Benchmark {
  /* index = 1 .. rowsize;
     intmatrix = array [index,index] of integer;
   */

  public static final int rowsize = 40;

  int ma[][] = new int[rowsize+1][rowsize+1];
  int mb[][] = new int[rowsize+1][rowsize+1];
  int mr[][] = new int[rowsize+1][rowsize+1];

  void initMatrix(int m[][]) {
    for (int i = 1; i <= rowsize; i++ )
      for (int j = 1; j <= rowsize; j++ ) {
        int temp = randomInt();
        m[i][j] = temp - (temp/120)*120 - 60;
      }
  };

  /* computes the inner product of A[row,*] and B[*,column] */
  int innerproduct(int a[][], int b[][], int row, int column) {
    // cached arow instead of a[row] (huge speed boost there)
    int result = 0;
    int arow[] = a[row];

    for(int i = 1; i <= rowsize; i++)
      result += arow[i] * b[i][column];

    return result;
  };

  public void run() {
    initRandom();
    initMatrix(ma);
    initMatrix(mb);
    for (int i = 1; i <= rowsize; i++ )
      for (int j = 1; j <= rowsize; j++ ) 
        mr[i][j] = innerproduct(ma,mb,i,j);
  };
}

final class DoubleMatrixMultiply extends Benchmark {
  /* index = 1 .. rowsize;
     intmatrix = array [index,index] of integer;
   */

  public static final int rowsize = 40;

  double ma[][] = new double[rowsize+1][rowsize+1];
  double mb[][] = new double[rowsize+1][rowsize+1];
  double mr[][] = new double[rowsize+1][rowsize+1];

  void initMatrix(double m[][]) {
    int temp, i, j;
    for ( i = 1; i <= rowsize; i++ )
      for ( j = 1; j <= rowsize; j++ ) {
        temp = randomInt();
        // JAM: should this be /3d? it's all int arithmetic
        m[i][j] = (temp - (temp/120)*120 - 60)/3;
      }
  };

  /* computes the inner product of A[row,*] and B[*,column] */
  double innerproduct(double a[][], double b[][], int row, int column) {
    // cached arow instead of a[row] (huge speed boost there)
    double result = 0;
    double arow[] = a[row];

    for(int i = 1; i <= rowsize; i++)
      result += arow[i] * b[i][column];

    return result;
  };

  public void run() {
    initRandom();
    initMatrix(ma);
    initMatrix(mb);
    for (int i = 1; i <= rowsize; i++ )
      for (int j = 1; j <= rowsize; j++ ) 
        mr[i][j] = innerproduct(ma,mb,i,j);
  };
}


final class Puzzle extends Benchmark {
  /* A compute-bound program from Forest Baskett. */
  /*
     piececlass = 0..classmax;
     piecetype = 0..typemax;
     position = 0..size;
   */
  private static final int size = 511;
  private static final int classmax = 3;
  private static final int typemax = 12;
  private static final int d = 8;

  int piececount[] = new int[classmax+1];
  int type[] = new int[typemax+1];
  int piecemax[] = new int[typemax+1];
  boolean puzzl[] = new boolean[size+1];
  boolean p[][] = new boolean[typemax+1][size+1];
  int n;
  int kount;

  boolean fit(int i, int j) {
    boolean pi[] = p[i];
    int piecemaxi = piecemax[i];

    for (int k = 0; k <= piecemaxi; k++ ) {
      if ( pi[k]) if (puzzl[j+k] ) 
        return false;
    }
    return true;
  };

  int place(int i, int j) {
    boolean pi[] = p[i];
    int piecemaxi = piecemax[i];

    for (int k = 0; k <= piecemaxi; k++ )
      if ( pi[k] ) puzzl[j+k] = true;

    piececount[type[i]] = piececount[type[i]] - 1;
    for (int k = j; k <= size; k++ )
      if ( !puzzl[k] ) {
        return (k);
      };
    return (0);
  };

  void remove(int i, int j) {
    boolean pi[] = p[i];
    int piecemaxi = piecemax[i];

    for (int k = 0; k <= piecemaxi; k++ )
      if ( pi[k] ) puzzl[j+k] = false;
    piececount[type[i]] = piececount[type[i]] + 1;
  };

  boolean trial(int j) {
    int i, k;
    kount = kount + 1;
    for ( i = 0; i <= typemax; i++ )
      if ( piececount[type[i]] != 0 )
        if ( fit(i, j) ) {
          k = place(i, j);
          if ( trial(k) || (k == 0) ) {
            return true;
          }
          else remove(i, j);
        };
    return false;
  };

  public void run() {
    int i, j, k, m;
    for ( m = 0; m <= size; m++ ) puzzl[m] = true;
    for( i = 1; i <= 5; i++ )for( j = 1; j <= 5; j++ )for( k = 1; k <= 5; k++ )
      puzzl[i+d*(j+d*k)] = false;
    for( i = 0; i <= typemax; i++ )for( m = 0; m<= size; m++ ) p[i][m] = false;

    for( i = 0; i <= 3; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 0; k++ )
      p[0][i+d*(j+d*k)] = true;
    type[0] = 0;
    piecemax[0] = 3+d*1+d*d*0;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 3; k++ )
      p[1][i+d*(j+d*k)] = true;
    type[1] = 0;
    piecemax[1] = 1+d*0+d*d*3;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 3; j++ )for( k = 0; k <= 1; k++ )
      p[2][i+d*(j+d*k)] = true;
    type[2] = 0;
    piecemax[2] = 0+d*3+d*d*1;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 3; j++ )for( k = 0; k <= 0; k++ )
      p[3][i+d*(j+d*k)] = true;
    type[3] = 0;
    piecemax[3] = 1+d*3+d*d*0;

    for( i = 0; i <= 3; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 1; k++ )
      p[4][i+d*(j+d*k)] = true;
    type[4] = 0;
    piecemax[4] = 3+d*0+d*d*1;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 3; k++ )
      p[5][i+d*(j+d*k)] = true;
    type[5] = 0;
    piecemax[5] = 0+d*1+d*d*3;

    for( i = 0; i <= 2; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 0; k++ )
      p[6][i+d*(j+d*k)] = true;
    type[6] = 1;
    piecemax[6] = 2+d*0+d*d*0;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 2; j++ )for( k = 0; k <= 0; k++ )
      p[7][i+d*(j+d*k)] = true;
    type[7] = 1;
    piecemax[7] = 0+d*2+d*d*0;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 2; k++ )
      p[8][i+d*(j+d*k)] = true;
    type[8] = 1;
    piecemax[8] = 0+d*0+d*d*2;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 0; k++ )
      p[9][i+d*(j+d*k)] = true;
    type[9] = 2;
    piecemax[9] = 1+d*1+d*d*0;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 1; k++ )
      p[10][i+d*(j+d*k)] = true;
    type[10] = 2;
    piecemax[10] = 1+d*0+d*d*1;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 1; k++ )
      p[11][i+d*(j+d*k)] = true;
    type[11] = 2;
    piecemax[11] = 0+d*1+d*d*1;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 1; k++ )
      p[12][i+d*(j+d*k)] = true;
    type[12] = 3;
    piecemax[12] = 1+d*1+d*d*1;

    piececount[0] = 13;
    piececount[1] = 3;
    piececount[2] = 1;
    piececount[3] = 1;
    m = 1+d*(1+d*1);
    kount = 0;
    if ( fit(0, m) ) n = place(0, m);
    else System.out.println("Error1 in Puzzle");
    if ( !trial(n) ) System.out.println ("Error2 in Puzzle.");
    else if ( kount != 2005 ) System.out.println ( "Error3 in Puzzle.");
  };


}

abstract class Sort extends Benchmark {
  /* listsize = 0..sortelements;
     sortarray = array [listsize] of integer;
   */
  public static final int sortelements = 5000;
  public static final int bsortelements = 500;

  int sortlist[] = new int[sortelements+1];
  int biggest, littlest, top;

  void initArray(int size) {
    int i;
    initRandom();
    biggest = 0; littlest = 0;
    for ( i = 1; i <= size; i++ )
    {
      sortlist[i] = randomInt();
      if ( sortlist[i] > biggest ) biggest = sortlist[i];
      else if ( sortlist[i] < littlest ) littlest = sortlist[i];
    };
  };

}

final class QuickSort extends Sort {
  void sort(int a[], int l, int r) {
    int i,j,x,w;

    i=l; j=r;
    x=a[(l+r) / 2];
    do {
      int ai = a[i];
      int aj = a[j];

      while ( ai<x ) ai=a[++i];
      //while ( a[i]<x ) i++;

      while ( x<aj ) aj=a[--j];
      //while ( x<a[j] ) j--;

      if ( i<=j ) {
        a[i] = aj;
        a[j] = ai;
        i++; j--;
      }
    } while ( i<=j );
    if ( l<j ) sort(a,l,j);
    if ( i<r ) sort(a,i,r);
  };

  public void run() {
    initArray(sortelements);
    sort(sortlist,1,sortelements);
    //Arrays.sort(sortlist,1,sortelements+1); // slightly slower quicksort
    if ( (sortlist[1] != littlest) || (sortlist[sortelements] != biggest) )
      System.out.println ( " Error in Quick.");
  };
}

final class BubbleSort extends Sort {

  public void run() {
    initArray(bsortelements);

    // tried to remove array bounds checks; origial C used while
    // loops which might be harder to optimize?, and didn't store
    // i+1 or j,k values.  

    for (int top=bsortelements; top>1; top--) {
      for (int i=1; i < top; i++) {
        final int nextIndex = i+1;
        final int j = sortlist[i];
        final int k = sortlist[nextIndex];

        if ( j > k ) {
          sortlist[i] = k;
          sortlist[nextIndex] = j;
        };
      }
    }

    if ( (sortlist[1] != littlest) || (sortlist[bsortelements] != biggest) )
      System.out.println ( "Error3 in Bubble.");
  }
}

final class TreeSort extends Sort {
  final static class Node {
    Node(int val) {
      this.val = val;
    }
    Node left, right;
    int val;
  };

  Node tree;

  void insert(int n, Node t) {
    if ( n > t.val ) 
      if ( t.left == null ) t.left = new Node(n);
      else insert(n,t.left);
    else if ( n < t.val )
      if ( t.right == null ) t.right = new Node(n);
      else insert(n,t.right);
  };

  boolean checkTree(Node p) {
    boolean result = true;

    if (p.left != null) 
      if (p.left.val <= p.val) 
        result = false;
      else 
        result = checkTree(p.left) && result;

    if (p.right != null)
      if (p.right.val >= p.val) 
        result = false;
      else 
        result = checkTree(p.right) && result;

    return result;
  };

  void deleteTree(Node p) {
    if ( p.left != null ) deleteTree(p.left);
    if ( p.right != null ) deleteTree(p.right);

    // free(p);
  }

  public void run() {
    initArray(sortelements);

    tree = new Node(sortlist[1]);

    for (int i = 2; i <= sortelements; i++ ) 
      insert(sortlist[i],tree);

    if (!checkTree(tree)) System.out.println ( " Error in Tree.");

    //deleteTree(tree);
  };
}


final class FFT extends Benchmark {
  /* carray = array [1..fftsize] of complex ;
     c2array = array [1..fftsize2] of complex ;
   */
  public static final int fftsize = 256;
  public static final int fftsize2 = 129;

  // much faster than CompexArray rp[]/ip[]
  final static class Complex {
    double rp;
    double ip; 
  };

  Complex z[] = new Complex[fftsize+1];
  Complex w[] = new Complex[fftsize+1];
  Complex e[] = new Complex[fftsize2+1];

  double    zr, zi;


  /* computes cos of x (x in radians) by an expansion */
  double cos(double x) {
    int i, factor;
    double result, power;

    result = 1.0; factor = 1; power = x;
    for ( i = 2; i <= 10; i++ ) {
      factor = factor * i;  power = power*x;
      if ( (i & 1) == 0 )  {
        if ( (i & 3) == 0 ) result = result + power/factor;
        else result = result - power/factor;
      };
    };
    return result;
  };

  int min(int arg1, int arg2) {
    if ( arg1 < arg2 )
      return (arg1);
    else
      return (arg2);
  };

  void printComplex(int arg1, int arg2, Complex zarray[], int start, int finish, int increment) {
    int i;
    System.out.println("");

    i = start;
    do {
      System.out.println("  "+zarray[i].rp+" "+zarray[i].ip); // format was %15.3e
      i = i + increment;
      System.out.println("  "+zarray[i].rp+" "+zarray[i].ip); // format was %15.3e
      System.out.println("");
      i = i + increment ;
    } while ( i <= finish );
  } ;

  /* original C version
     Uniform11(iy, yfl) int *iy; double *yfl;
     {
   *iy = (4855*(*iy) + 1731) & 8191;
   *yfl = (*iy)/8192.0;
   } 
   */

  int uniform11_iy(int iy) {
    return (4855*(iy) + 1731) & 8191;
  };
  double uniform11_yfl(int iy) {
    return (iy)/8192.0;
  }

  void exptab(int n, Complex e[]) { /* exptab */
    double theta, divisor;
    double h[] = new double[26];
    int i, j, k, l, m;

    theta = 3.1415926536;
    divisor = 4.0;
    for (i=1; i <= 25; i++) {
      h[i] = 1.0/(2.0*cos( theta/divisor ));
      divisor = divisor + divisor;
    };

    m = n / 2 ;
    l = m / 2 ;
    j = 1 ;
    e[1].rp = 1.0 ;
    e[1].ip = 0.0;
    e[l+1].rp = 0.0;
    e[l+1].ip = 1.0 ;
    e[m+1].rp = -1.0 ;
    e[m+1].ip = 0.0 ;

    do {
      i = l / 2 ;
      k = i ;

      do {
        e[k+1].rp = h[j]*(e[k+i+1].rp+e[k-i+1].rp) ;
        e[k+1].ip = h[j]*(e[k+i+1].ip+e[k-i+1].ip) ;
        k = k+l ;
      } while ( k <= m );

      j = min( j+1, 25);
      l = i ;
    } while ( l > 1 );

  } /* exptab */ ;

  void fft(int n, Complex z[], Complex w[], Complex e[], double sqrinv) {
    int i, j, k, l, m, index;
    m = n / 2 ;
    l = 1 ;

    do {
      k = 0 ;
      j = l ;
      i = 1 ;

      do {
        do {
          Complex wik = w[i+k];
          Complex wij = w[i+j];
          Complex zmi = z[m+i];
          Complex zi = z[i];
          Complex ek1 = e[k+1];

          wik.rp = zi.rp+zmi.rp ;
          wik.ip = zi.ip+zmi.ip ;
          wij.rp = ek1.rp*(zi.rp-zmi.rp)
                  -ek1.ip*(zi.ip-zmi.ip) ;
          wij.ip = ek1.rp*(zi.ip-zmi.ip)
                  +ek1.ip*(zi.rp-zmi.rp) ;
          i = i+1 ;
        } while ( i <= j );

        k = j ;
        j = k+l ;
      } while ( j <= m );

      /*z = w ;*/ index = 1;
      do {
        z[index].rp = w[index].rp;
        z[index].ip = w[index].ip;
        index = index+1;
      } while ( index <= n );
      l = l+l ;
    } while ( l <= m );

    for ( i = 1; i <= n; i++ )
    {
      z[i].rp = sqrinv*z[i].rp ;
      z[i].ip = -sqrinv*z[i].ip;
    } ;

  } ;

  public void run() { /* oscar */
    for (int i=0; i < fftsize+1; i++) {
      z[i] = new Complex();
      w[i] = new Complex();
    }
    for (int i=0; i < fftsize2+1; i++) {
      e[i] = new Complex();
    }

    exptab(fftsize,e) ;
    seed = 5767 ;

    for (int i = 1; i <= fftsize; i++ ) {
      seed = uniform11_iy(seed);
      zr = uniform11_yfl(seed);

      seed = uniform11_iy(seed);
      zi = uniform11_yfl(seed);

      z[i].rp = 20.0*zr - 10.0;
      z[i].ip = 20.0*zi - 10.0;
    } ;

    for (int i = 1; i <= 20; i++ ) {
      fft(fftsize,z,w,e,0.0625) ;
    };

    if (   z[1].rp < -9.3653 || z[1].rp > -9.3652 
        || z[1].ip < -3.9869 || z[1].ip > -3.9868)
      System.out.println ( " Error in FFT.");

  } /* oscar */ ;

}



final class Perm extends Benchmark {
  public static final int permrange = 10;

  final int permarray[] = new int[permrange+1];
  int pctr;

  void initialize() {
    for (int i=1; i <= 7; i++ ) {
      permarray[i]=i-1;
    }
  };

  void permute(int n) {   /* permute */
    pctr = pctr + 1;
    if (n!=1) {
      permute(n-1);
      for (int k = n-1; k >= 1; k--) {
        int tmp;

        // was call to Swap but impossible in Java
        tmp = permarray[n];
        permarray[n] = permarray[k];
        permarray[k] = tmp;

        permute(n-1);

        tmp = permarray[n];
        permarray[n] = permarray[k];
        permarray[k] = tmp;
      }
    }
  }

  public void run() {   /* Perm */
    pctr = 0;
    for (int i = 1; i <= 5; i++ ) {
      initialize();
      permute(7);
    };
    if ( pctr != 43300 )
      System.out.println(" Error in Perm.\n");
  }
}




final class Queen extends Benchmark {
  /* The eight queens problem, solved 50 times. */
  /*
     type    
     doubleboard =   2..16;
     doublenorm  =   -7..7;
     boardrange  =   1..8;
     aarray      =   array [boardrange] of boolean;
     barray      =   array [doubleboard] of boolean;
     carray      =   array [doublenorm] of boolean;
     xarray      =   array [boardrange] of boardrange;
   */

  // the C hackers passed *q as a global exit flag int, so in
  // this version it's a returned boolean since there's no pass
  // by reference in Java or if(int).  why a,b,c aren't
  // boolean???

  boolean Try(int i, boolean q, boolean a[], boolean b[], boolean c[], int x[]) {
    q = false;
    for (int j=1; j <= 8 && !q; j++) {
      int ij = i+j;
      int ij7 = i-j+7;

      if ( b[j] && a[ij] && c[ij7] ) {
        x[i] = j;
        b[j] = false;
        a[ij] = false;
        c[ij7] = false;
        if ( i < 8 ) {
          q=Try(i+1,q,a,b,c,x);
          if ( ! q ) {
            b[j] = true;
            a[ij] = true;
            c[ij7] = true;
          }
        }
        else q = true;
      }
    }

    return q;
  };

  void doit() {
    int i;
    boolean q=false;
    boolean a[] = new boolean[9];
    boolean b[] = new boolean[17];
    boolean c[] = new boolean[15];
    int x[] = new int[9];

    i = 0 - 7;
    while ( i <= 16 ) { 
      if ( (i >= 1) && (i <= 8) ) a[i] = true;
      if ( i >= 2 ) b[i] = true;
      if ( i <= 7 ) c[i+7] = true;
      i = i + 1;
    };

    q=Try(1, q, b, a, c, x);
    if ( ! q )
      System.out.println (" Error in Queens.");
  };

  public void run() { /* queens */
    for (int i = 1; i <= 50; i++ ) doit();
  };

}

