
import java.util.*;

/**
 * This is a suite of benchmarks that are relatively short, both
 * in program size and execution time.  It requires no input, and
 * prints the execution time for each program.  It does a
 * rudimentary check to make sure each program gets the right
 * output.<p>
 *
 * These programs were gathered by John Hennessy and modified by
 * Peter Nye and ported to Java by Jon Maxwell.<p>
 */

// GVG (Greg Gritton): Will need to increase ITER [JAM: above 500?]


public final class Bench {

  public static final int ITER = 500;

  public static final double bubblebase = 1.61;
  public static final double dnfbase = 3.5;
  public static final double permbase = 1.75;
  public static final double queensbase = 1.83;
  public static final double towersbase = 2.39;
  public static final double quickbase = 1.92;
  public static final double intmmbase = 1.46;
  public static final double treebase = 2.5;
  public static final double mmbase = 0.0 /* 0.73 */;
  public static final double fpmmbase = 2.92;
  public static final double puzzlebase = 0.5;
  public static final double fftbase = 0.0 /* 1.11 */;
  public static final double fpfftbase = 4.44;

  static double value;
  static double fixed, floated, total;
  static double st_total; /* Total time for same # of iterations as Smalltalk runs */

  static String format(double value, int places, int decimals) {
    return ""+value; // fix to use formatter
  }

  static void runTest(String name, Benchmark benchmark, double fixed_base, double float_base, int st_base) {
    long starttime; /* starting time for test */
    double runtime; /* total time the test runs */

    System.gc();
    starttime = System.currentTimeMillis();
    for (int i=0; i<ITER; i++) {
      benchmark.run();
    }
    runtime = (System.currentTimeMillis()-starttime)/(double)ITER;
    System.gc();

    System.out.println(name+":    "+format(runtime,6,2));
    fixed += fixed_base * runtime;
    floated += float_base * runtime;
    st_total += st_base * runtime;
    total += runtime;
  }

  public static void main(String args[]) {
    fixed = floated = st_total = total = 0.0;

    System.out.println("Running "+ITER+" iterations of each benchmark.\n");
    System.out.println("Reporting timer per iteration in milliseconds.\n");

    /* output results */
    runTest("  Bubble", new BubbleSort(), bubblebase, bubblebase, 10);
    runTest("   Intmm", new IntMatrixMultiply() , intmmbase , intmmbase , 10);
    runTest("      Mm", new DoubleMatrixMultiply()    , mmbase    , fpmmbase  , 10);
    runTest("    Perm", new Perm()  , permbase  , permbase  , 10);
    runTest("  Queens", new Queen(), queensbase, queensbase, 10);
    runTest("   Quick", new QuickSort() , quickbase , quickbase , 10);
    runTest("    Tree", new TreeSort() , treebase  , treebase  , 10);
    runTest("  Towers", new Towers(), towersbase, towersbase, 10);
    runTest("     FFT", new FFT() , fftbase   , fpfftbase ,  5);
    runTest("  Puzzle", new Puzzle(), puzzlebase, puzzlebase,  3);

    /* compute composites */
    System.out.println("");
    System.out.println("Smalltalk runSome equivalent total is " + format(st_total, 5, 0));
    System.out.println("Simple total is                       " + format(total, 6, 1));
    System.out.println("Nonfloating point composite is        " + format(fixed / ITER, 7, 2));
    System.out.println("Floating point composite is           " + format(floated / ITER, 7, 2));
    System.out.println("");
  }

}


abstract class Benchmark {
  public static final int falseInt = 0;
  public static final int trueInt = 1;
  static int seed;

  static void initRandom() {
    seed = 74755;
  };

  static int randomInt() {
    seed = (seed * 1309 + 13849) & 65535;
    return seed;
  };

  abstract public void run();
};










class Towers extends Benchmark {
  public static final int maxcells = 18;
  public static final int stackrange = 3;

  /* discsizrange = 1..maxcells; */
  /*    cellcursor = 0..maxcells; */

  static final class Element {
    int discsize;
    int next;
  };

  int freelist, movesdone;
  int stack[] = new int[stackrange+1];
  Element cellspace[] = new Element[maxcells+1];
  {
    for (int i=0; i < maxcells+1; i++)
      cellspace[i] = new Element();
  };

  void error(String emsg) {
    System.out.println(" Error in Towers: "+emsg);
  };

  void makenull(int s) {
    stack[s]=0;
  };

  int getElement() {
    int temp = 0;
    if ( freelist>0 ) {
      temp = freelist;
      freelist = cellspace[freelist].next;
    }
    else
      error("out of space   ");
    return temp;
  };

  void push(int i,int s) {
    boolean errorfound;
    int localel;

    errorfound=false;
    if ( stack[s] > 0 )
      if ( cellspace[stack[s]].discsize<=i ) {
        errorfound=true;
        error("disc size error");
      };
    if ( !errorfound ) {
      localel=getElement();
      cellspace[localel].next=stack[s];
      stack[s]=localel;
      cellspace[localel].discsize=i;
    }
  };

  void init(int s, int n) {
    int discctr;
    makenull(s);
    for ( discctr = n; discctr >= 1; discctr-- )
      push(discctr,s);
  };

  int pop(int s) {
    int temp, temp1;
    if ( stack[s] > 0 ) {
      temp1 = cellspace[stack[s]].discsize;
      temp = cellspace[stack[s]].next;
      cellspace[stack[s]].next=freelist;
      freelist=stack[s];
      stack[s]=temp;
      return temp1;
    }
    else
      error("nothing to pop ");
    return 0;
  };

  void move(int s1, int s2) {
    push(pop(s1),s2);
    movesdone=movesdone+1;
  };

  void tower(int i, int j, int k) {
    int other;
    if ( k==1 )
      move(i,j);
    else {
      other=6-i-j;
      tower(i,other,k-1);
      move(i,j);
      tower(other,j,k-1);
    }
  };

  public void run() {
    int i;
    for ( i=1; i <= maxcells; i++ )
      cellspace[i].next=i-1;
    freelist=maxcells;
    init(1,14);
    makenull(2);
    makenull(3);
    movesdone=0;
    tower(1,2,14);
    if ( movesdone != 16383 )
      System.out.println (" Error in Towers.");
  };
}

class IntMatrixMultiply extends Benchmark {
  /* index = 1 .. rowsize;
     intmatrix = array [index,index] of integer;
   */

  public static final int rowsize = 40;

  int ma[][] = new int[rowsize+1][rowsize+1];
  int mb[][] = new int[rowsize+1][rowsize+1];
  int mr[][] = new int[rowsize+1][rowsize+1];

  void initMatrix(int m[][]) {
    int temp, i, j;
    for ( i = 1; i <= rowsize; i++ )
      for ( j = 1; j <= rowsize; j++ ) {
        temp = randomInt();
        m[i][j] = temp - (temp/120)*120 - 60;
      }
  };

  /* computes the inner product of A[row,*] and B[*,column] */
  void innerproduct(int result[][], int a[][], int b[][], int row, int column) {
    result[row][column] = 0;
    for(int i = 1; i <= rowsize; i++)
      result[row][column] += a[row][i] * b[i][column];
  };

  public void run() {
    int i, j;
    initRandom();
    initMatrix(ma);
    initMatrix(mb);
    for ( i = 1; i <= rowsize; i++ )
      for ( j = 1; j <= rowsize; j++ ) 
        innerproduct(mr,ma,mb,i,j);
  };
}

class DoubleMatrixMultiply extends Benchmark {
  /* index = 1 .. rowsize;
     intmatrix = array [index,index] of integer;
   */

  public static final int rowsize = 40;

  double ma[][] = new double[rowsize+1][rowsize+1];
  double mb[][] = new double[rowsize+1][rowsize+1];
  double mr[][] = new double[rowsize+1][rowsize+1];

  void initMatrix(double m[][]) {
    int temp, i, j;
    for ( i = 1; i <= rowsize; i++ )
      for ( j = 1; j <= rowsize; j++ ) {
        temp = randomInt();
        // JAM: should this be /3d? it's all int arithmetic
        m[i][j] = (temp - (temp/120)*120 - 60)/3;
      }
  };

  /* computes the inner product of A[row,*] and B[*,column] */
  void innerproduct(double result[][], double a[][], double b[][], int row, int column) {
    result[row][column] = 0;
    for(int i = 1; i <= rowsize; i++ )
      result[row][column] += a[row][i] * b[i][column];
  };

  public void run() {
    int i, j;
    initRandom();
    initMatrix(ma);
    initMatrix(mb);
    for ( i = 1; i <= rowsize; i++ )
      for ( j = 1; j <= rowsize; j++ ) 
        innerproduct(mr,ma,mb,i,j);
  };
}


class Puzzle extends Benchmark {
  /* A compute-bound program from Forest Baskett. */
  /*
     piececlass = 0..classmax;
     piecetype = 0..typemax;
     position = 0..size;
   */
  public static final int size = 511;
  public static final int classmax = 3;
  public static final int typemax = 12;
  public static final int d = 8;

  int piececount[] = new int[classmax+1];
  int type[] = new int[typemax+1];
  int piecemax[] = new int[typemax+1];
  int puzzl[] = new int[size+1];
  int p[][] = new int[typemax+1][size+1];
  int n;
  int kount;

  int fit(int i, int j) {
    for (int k = 0; k <= piecemax[i]; k++ )
      if ( p[i][k]!=0 ) if ( puzzl[j+k]!=0 ) return (falseInt);
    return (trueInt);
  };

  int place(int i, int j) {
    int k;
    for ( k = 0; k <= piecemax[i]; k++ )
      if ( p[i][k]!=0 ) puzzl[j+k] = trueInt;
    piececount[type[i]] = piececount[type[i]] - 1;
    for ( k = j; k <= size; k++ )
      if ( puzzl[k]==0 ) {
        return (k);
      };
    return (0);
  };

  void remove(int i, int j) {
    int k;
    for ( k = 0; k <= piecemax[i]; k++ )
      if ( p[i][k]!=0 ) puzzl[j+k] = falseInt;
    piececount[type[i]] = piececount[type[i]] + 1;
  };

  int trial(int j) {
    int i, k;
    kount = kount + 1;
    for ( i = 0; i <= typemax; i++ )
      if ( piececount[type[i]] != 0 )
        if ( fit(i, j)!=0 ) {
          k = place(i, j);
          if ( trial(k)!=0 || (k == 0) ) {
            return (trueInt);
          }
          else remove(i, j);
        };
    return (falseInt);
  };

  public void run() {
    int i, j, k, m;
    for ( m = 0; m <= size; m++ ) puzzl[m] = trueInt;
    for( i = 1; i <= 5; i++ )for( j = 1; j <= 5; j++ )for( k = 1; k <= 5; k++ )
      puzzl[i+d*(j+d*k)] = falseInt;
    for( i = 0; i <= typemax; i++ )for( m = 0; m<= size; m++ ) p[i][m] = falseInt;

    for( i = 0; i <= 3; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 0; k++ )
      p[0][i+d*(j+d*k)] = trueInt;
    type[0] = 0;
    piecemax[0] = 3+d*1+d*d*0;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 3; k++ )
      p[1][i+d*(j+d*k)] = trueInt;
    type[1] = 0;
    piecemax[1] = 1+d*0+d*d*3;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 3; j++ )for( k = 0; k <= 1; k++ )
      p[2][i+d*(j+d*k)] = trueInt;
    type[2] = 0;
    piecemax[2] = 0+d*3+d*d*1;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 3; j++ )for( k = 0; k <= 0; k++ )
      p[3][i+d*(j+d*k)] = trueInt;
    type[3] = 0;
    piecemax[3] = 1+d*3+d*d*0;

    for( i = 0; i <= 3; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 1; k++ )
      p[4][i+d*(j+d*k)] = trueInt;
    type[4] = 0;
    piecemax[4] = 3+d*0+d*d*1;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 3; k++ )
      p[5][i+d*(j+d*k)] = trueInt;
    type[5] = 0;
    piecemax[5] = 0+d*1+d*d*3;

    for( i = 0; i <= 2; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 0; k++ )
      p[6][i+d*(j+d*k)] = trueInt;
    type[6] = 1;
    piecemax[6] = 2+d*0+d*d*0;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 2; j++ )for( k = 0; k <= 0; k++ )
      p[7][i+d*(j+d*k)] = trueInt;
    type[7] = 1;
    piecemax[7] = 0+d*2+d*d*0;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 2; k++ )
      p[8][i+d*(j+d*k)] = trueInt;
    type[8] = 1;
    piecemax[8] = 0+d*0+d*d*2;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 0; k++ )
      p[9][i+d*(j+d*k)] = trueInt;
    type[9] = 2;
    piecemax[9] = 1+d*1+d*d*0;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 0; j++ )for( k = 0; k <= 1; k++ )
      p[10][i+d*(j+d*k)] = trueInt;
    type[10] = 2;
    piecemax[10] = 1+d*0+d*d*1;

    for( i = 0; i <= 0; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 1; k++ )
      p[11][i+d*(j+d*k)] = trueInt;
    type[11] = 2;
    piecemax[11] = 0+d*1+d*d*1;

    for( i = 0; i <= 1; i++ )for( j = 0; j <= 1; j++ )for( k = 0; k <= 1; k++ )
      p[12][i+d*(j+d*k)] = trueInt;
    type[12] = 3;
    piecemax[12] = 1+d*1+d*d*1;

    piececount[0] = 13;
    piececount[1] = 3;
    piececount[2] = 1;
    piececount[3] = 1;
    m = 1+d*(1+d*1);
    kount = 0;
    if ( fit(0, m)!=0 ) n = place(0, m);
    else System.out.println("Error1 in Puzzle");
    if ( trial(n)==0 ) System.out.println ("Error2 in Puzzle.");
    else if ( kount != 2005 ) System.out.println ( "Error3 in Puzzle.");
  };


}

abstract class Sort extends Benchmark {
  /* listsize = 0..sortelements;
     sortarray = array [listsize] of integer;
   */
  public static final int sortelements = 5000;
  public static final int srtelements = 500;

  int sortlist[] = new int[sortelements+1];
  int biggest, littlest, top;

  void initArray() {
    int i, temp;
    initRandom();
    biggest = 0; littlest = 0;
    for ( i = 1; i <= sortelements; i++ )
    {
      temp = randomInt();
      sortlist[i] = temp - (temp/100000)*100000 - 50000;
      if ( sortlist[i] > biggest ) biggest = sortlist[i];
      else if ( sortlist[i] < littlest ) littlest = sortlist[i];
    };
  };

}

class QuickSort extends Sort {
  void sort(int a[], int l, int r) {
    int i,j,x,w;

    i=l; j=r;
    x=a[(l+r) / 2];
    do {
      while ( a[i]<x ) i = i+1;
      while ( x<a[j] ) j = j-1;
      if ( i<=j ) {
        w = a[i];
        a[i] = a[j];
        a[j] = w;
        i = i+1;    j= j-1;
      }
    } while ( i<=j );
    if ( l <j ) sort(a,l,j);
    if ( i<r ) sort(a,i,r);
  };

  public void run() {
    initArray();
    sort(sortlist,1,sortelements);
    if ( (sortlist[1] != littlest) || (sortlist[sortelements] != biggest) )
      System.out.println ( " Error in Quick.");
  };
}

class BubbleSort extends Sort {
  void initArray() {
    int i, temp;
    initRandom();
    biggest = 0; littlest = 0;
    for ( i = 1; i <= srtelements; i++ )
    {
      temp = randomInt();
      sortlist[i] = temp - (temp/100000)*100000 - 50000;
      if ( sortlist[i] > biggest ) biggest = sortlist[i];
      else if ( sortlist[i] < littlest ) littlest = sortlist[i];
    };
  };

  public void run() {
    int i, j;
    initArray();
    top=srtelements;

    while ( top>1 ) {
      i=1;
      while ( i<top ) {

        if ( sortlist[i] > sortlist[i+1] ) {
          j = sortlist[i];
          sortlist[i] = sortlist[i+1];
          sortlist[i+1] = j;
        };
        i=i+1;
      };

      top=top-1;
    };

    if ( (sortlist[1] != littlest) || (sortlist[srtelements] != biggest) )
      System.out.println ( "Error3 in Bubble.");
  }
}

class TreeSort extends Sort {
  class Node {
    Node left, right;
    int val;
  };
  Node tree;

  Node createNode(int n) {
    Node t = new Node();
    t.left = null; 
    t.right = null;
    t.val = n;
    return t;
  };

  void insert(int n, Node t) {
    if ( n > t.val ) 
      if ( t.left == null ) t.left = createNode(n);
      else insert(n,t.left);
    else if ( n < t.val )
      if ( t.right == null ) t.right = createNode(n);
      else insert(n,t.right);
  };

  boolean checkTree(Node p) {
    boolean result = true;

    if (p.left != null) 
      if (p.left.val <= p.val) 
        result = false;
      else 
        result = checkTree(p.left) && result;

    if (p.right != null)
      if (p.right.val >= p.val) 
        result = false;
      else 
        result = checkTree(p.right) && result;

    return result;
  };

  void deleteTree(Node p) {
    if ( p.left != null ) deleteTree(p.left);
    if ( p.right != null ) deleteTree(p.right);

    // free(p);
  }

  public void run() {
    initArray();

    tree = new Node();
    tree.left = null; 
    tree.right = null; 
    tree.val = sortlist[1];

    for (int i = 2; i <= sortelements; i++ ) 
      insert(sortlist[i],tree);

    if (!checkTree(tree)) System.out.println ( " Error in Tree.");

    //deleteTree(tree);
    //System.gc();
  };
}


final class FFT extends Benchmark {
  /* carray = array [1..fftsize] of complex ;
     c2array = array [1..fftsize2] of complex ;
   */
  public static final int fftsize = 256;
  public static final int fftsize2 = 129;

  // complex stored as an array of struct { int, double } in the C version
  final static class Complex {
    double rp;
    double ip; 
  };

  Complex z[] = new Complex[fftsize+1];
  Complex w[] = new Complex[fftsize+1];
  Complex e[] = new Complex[fftsize2+1];

  double    zr, zi;


  /* computes cos of x (x in radians) by an expansion */
  double cos(double x) {
    int i, factor;
    double result, power;

    result = 1.0; factor = 1; power = x;
    for ( i = 2; i <= 10; i++ ) {
      factor = factor * i;  power = power*x;
      if ( (i & 1) == 0 )  {
        if ( (i & 3) == 0 ) result = result + power/factor;
        else result = result - power/factor;
      };
    };
    return result;
  };

  int min(int arg1, int arg2) {
    if ( arg1 < arg2 )
      return (arg1);
    else
      return (arg2);
  };

  void printComplex(int arg1, int arg2, Complex zarray[], int start, int finish, int increment) {
    int i;
    System.out.println("");

    i = start;
    do {
      System.out.println("  "+zarray[i].rp+" "+zarray[i].ip); // format was %15.3e
      i = i + increment;
      System.out.println("  "+zarray[i].rp+" "+zarray[i].ip); // format was %15.3e
      System.out.println("");
      i = i + increment ;
    } while ( i <= finish );
  } ;

  /* original C version
     Uniform11(iy, yfl) int *iy; double *yfl;
     {
   *iy = (4855*(*iy) + 1731) & 8191;
   *yfl = (*iy)/8192.0;
   } 
   */

  int uniform11_iy(int iy) {
    return (4855*(iy) + 1731) & 8191;
  };
  double uniform11_yfl(int iy) {
    return (iy)/8192.0;
  }

  void exptab(int n, Complex e[]) { /* exptab */
    double theta, divisor;
    double h[] = new double[26];
    int i, j, k, l, m;

    theta = 3.1415926536;
    divisor = 4.0;
    for (i=1; i <= 25; i++) {
      h[i] = 1/(2*cos( theta/divisor ));
      divisor = divisor + divisor;
    };

    m = n / 2 ;
    l = m / 2 ;
    j = 1 ;
    e[1].rp = 1.0 ;
    e[1].ip = 0.0;
    e[l+1].rp = 0.0;
    e[l+1].ip = 1.0 ;
    e[m+1].rp = -1.0 ;
    e[m+1].ip = 0.0 ;

    do {
      i = l / 2 ;
      k = i ;

      do {
        e[k+1].rp = h[j]*(e[k+i+1].rp+e[k-i+1].rp) ;
        e[k+1].ip = h[j]*(e[k+i+1].ip+e[k-i+1].ip) ;
        k = k+l ;
      } while ( k <= m );

      j = min( j+1, 25);
      l = i ;
    } while ( l > 1 );

  } /* exptab */ ;

  void fft(int n, Complex z[], Complex w[], Complex e[], double sqrinv) {
    int i, j, k, l, m, index;
    m = n / 2 ;
    l = 1 ;

    do {
      k = 0 ;
      j = l ;
      i = 1 ;

      do {
        do {
          w[i+k].rp = z[i].rp+z[m+i].rp ;
          w[i+k].ip = z[i].ip+z[m+i].ip ;
          w[i+j].rp = e[k+1].rp*(z[i].rp-z[i+m].rp)
            -e[k+1].ip*(z[i].ip-z[i+m].ip) ;
          w[i+j].ip = e[k+1].rp*(z[i].ip-z[i+m].ip)
            +e[k+1].ip*(z[i].rp-z[i+m].rp) ;
          i = i+1 ;
        } while ( i <= j );

        k = j ;
        j = k+l ;
      } while ( j <= m );

      /*z = w ;*/ index = 1;
      do {
        z[index].rp = w[index].rp;
        z[index].ip = w[index].ip;
        index = index+1;
      } while ( index <= n );
      l = l+l ;
    } while ( l <= m );

    for ( i = 1; i <= n; i++ )
    {
      z[i].rp = sqrinv*z[i].rp ;
      z[i].ip = -sqrinv*z[i].ip;
    } ;

  } ;

  public void run() { /* oscar */
    // new complex objects
    for (int i=0; i < fftsize+1; i++) {
      z[i] = new Complex();
      w[i] = new Complex();
    }
    for (int i=0; i < fftsize2+1; i++) {
      e[i] = new Complex();
    }

    exptab(fftsize,e) ;
    seed = 5767 ;

    for (int i = 1; i <= fftsize; i++ ) {
      seed = uniform11_iy(seed);
      zr = uniform11_yfl(seed);

      seed = uniform11_iy(seed);
      zi = uniform11_yfl(seed);

      z[i].rp = 20.0*zr - 10.0;
      z[i].ip = 20.0*zi - 10.0;
    } ;

    for (int i = 1; i <= 20; i++ ) {
      fft(fftsize,z,w,e,0.0625) ;
      /* Printcomplex( 6, 99, z, 1, 256, 17 ); */
    };
  } /* oscar */ ;

}




class Perm extends Benchmark {
  public static final int permrange = 10;

  int permarray[] = new int[permrange+1];
  int pctr;

  void initialize() {
    for (int i=1; i <= 7; i++ ) {
      permarray[i]=i-1;
    }
  };

  void permute(int n) {   /* permute */
    int k;
    pctr = pctr + 1;
    if (n!=1) {
      permute(n-1);
      for (k = n-1; k >= 1; k--) {
        int tmp;

        // was call to Swap but impossible in Java
        tmp = permarray[n];
        permarray[n] = permarray[k];
        permarray[k] = tmp;

        permute(n-1);

        tmp = permarray[n];
        permarray[n] = permarray[k];
        permarray[k] = tmp;
      }
    }
  }

  public void run() {   /* Perm */
    pctr = 0;
    for (int i = 1; i <= 5; i++ ) {
      initialize();
      permute(7);
    };
    if ( pctr != 43300 )
      System.out.println(" Error in Perm.\n");
  }
}




class Queen extends Benchmark {
  /* The eight queens problem, solved 50 times. */
  /*
     type    
     doubleboard =   2..16;
     doublenorm  =   -7..7;
     boardrange  =   1..8;
     aarray      =   array [boardrange] of boolean;
     barray      =   array [doubleboard] of boolean;
     carray      =   array [doublenorm] of boolean;
     xarray      =   array [boardrange] of boardrange;
   */

  // the C hackers passed *q as a global exit flag int, so in
  // this version it's a returned boolean since there's no pass
  // by reference in Java or if(int).  why a,b,c aren't
  // boolean???

  boolean Try(int i, boolean q, int a[], int b[], int c[], int x[]) {
    int j;
    j = 0;
    q = false;
    while ( (! q) && (j != 8) ) { 
      j = j + 1;
      q = false;
      if ( b[j]!=0 && a[i+j]!=0 && c[i-j+7]!=0 ) {
        x[i] = j;
        b[j] = falseInt;
        a[i+j] = falseInt;
        c[i-j+7] = falseInt;
        if ( i < 8 ) {
          q=Try(i+1,q,a,b,c,x);
          if ( ! q ) {
            b[j] = trueInt;
            a[i+j] = trueInt;
            c[i-j+7] = trueInt;
          }
        }
        else q = true;
      }
    }

    return q;
  };

  void doit() {
    int i;
    boolean q=false;
    int a[] = new int[9];
    int b[] = new int[17];
    int c[] = new int[15];
    int x[] = new int[9];

    i = 0 - 7;
    while ( i <= 16 ) { 
      if ( (i >= 1) && (i <= 8) ) a[i] = trueInt;
      if ( i >= 2 ) b[i] = trueInt;
      if ( i <= 7 ) c[i+7] = trueInt;
      i = i + 1;
    };

    q=Try(1, q, b, a, c, x);
    if ( ! q )
      System.out.println (" Error in Queens.");
  };

  public void run() { /* queens */
    for (int i = 1; i <= 50; i++ ) doit();
  };

}

