"Filed out from Dolphin Smalltalk/Version 2.1"!

Object subclass: #BlockClosure
	instanceVariableNames: 'outer argumentCount initialIP method '
	classVariableNames: ''
	poolDictionaries: ''!

BlockClosure class instanceVariableNames: ''!

BlockClosure comment: ''!

!BlockClosure categoriesForClass!No category! !

!BlockClosure methodsFor!

argumentCount
	"Answer the <integer> number of arguments expected by the receiver."

	^argumentCount!

atPriority: anInteger
	"Evaluate the receiver at the specified priority. 
	On completion or curtailment the 	active process' 
	priority is returned to its previous value."

	| activePriority proc |
	proc := Processor activeProcess.
	activePriority := proc priority: anInteger.
	^self ensure: [proc priority: activePriority]!

critical
	"Private - Evaluate the receiver as a Critical Section with asynchronous process switching disabled.
	When asynchronous process switching is disabled, process pre-emption cannot occur because
	all interrupts and asynchronous Semaphore signals remain pending. If synchronous process
	synchronisation primitives are attempted (e.g. sending #signal or #wait to a Semaphore)
	then asynchronous process switching is automatically re-enabled, and any pending interrupts/
	asynchronous signals will get delivered.

	Use this VERY sparingly, and for very short durations, as this is a very broadbrush (though
	very high performance) approach to process synchronisation. It is an easy and high performance
	way to protect, for example, a shared data structure from being accessed by other processes
	while it is being modified, but it will prevent any other processes from executing unrelated
	code too. Also, it may not be available in a future multi-threaded VM.
	Synchronisation of access to shared data structures is best done by including a 
	mutual exclusion Semaphore or Mutex with the data structure (i.e. in the Object which manages the 
	data structure), and then using the #critical: method as here. 
	
	See the SharedQueue class for an example of the use of mutual exlusion semaphores, or SharedSet for
	and example of the use of a Mutex to guarantee mutually exclusive access.

	Implementation Note: We must use an #ensure: block to guarantee re-enabling of aysnc events, 
	because the receiver may perform a ^-return (should an exception occur then they will be 
	re-enabled anyway). This does mean it may be quicker to send #enableAsyncEvents: directly
	in some circumstances."

	| oldState |
	oldState := Processor enableAsyncEvents: false.
	[self value] ensure: [Processor enableAsyncEvents: oldState]!

ensure: terminationBlock
	"Evaluate the receiver, and regardless of the means by which it exits, 
	be it by normal or non-local (^) return, or by raising an exception, evaluate
	the niladic block argument <terminationBlock>.
	Answers the result of evaluating the receiver, unless <terminationBlock>
	contains a non-local return, in which case the result of <terminationBlock> will be
	answered to its home contexts sender, e.g.:

		[^1] ensure: [Sound beep]		answers 1, but also woofs
		[^1] ensure: [Sound beep. ^2]		woofs and answers 2.

	Be warned: The system may become unstable if this method is modified. The receiver is
	not the original receiver of the message after #valueOnUnwind: has been sent, but is
	valid inside the unwind block. In addition, #valueOnUnwind: leaves additional objects
	on the stack.

	See also #ifCurtailed:"

	| answer |
	answer := self valueOnUnwind: [:retValue :retFrame |
		terminationBlock value.
		Processor returnValue: retValue toFrame: retFrame].
	terminationBlock value.
	^answer!

fork
	"Create and schedule a new Process at the same priority as the current active process.
	The new Process will be placed on the end of the queue of processes of the same priority in 
	the Ready state (i.e. the current active process continues to run). Answer the new Process.
	The new Process will run when it gets its turn (i.e. when there are no higher
	priority processes in the Ready state, and the receiver is at the front of the
	queue of Ready Processes at its priority)."

	^self newProcess resume!

forkAt: anInteger
	"Evaluate the receiver in a new process at the priority specified by the argument, anInteger. 
	If the new process is of a higher priority than the current active process then it will
	start running immediately. If of a lower priority, then it must wait until there are no
	higher priority processes Ready to run. If of the same priority, then it does not preempt
	the current active process. N.B. Do not rely on the latter behaviour, as it may change in
	a future release.
	Answer the new process."

	^(self newProcess) 
		priority: anInteger;
		resume!

frameClass
	"Private - Answer the class of StackFrame to represent activations of the
	receiver."

	^BlockFrame!

home
	"Private - Answer the home context of the context. In the case of the BlockClosure,
	this is the MethodContext which represented the CompiledMethod which
	created the Block. N.B. The home context (which is always a MethodContext)
	may have already returned, and only exists in order to provide a 'home'.
	A BlockClosure uses its home to reference the receiver, method, and 
	its temporaries."

	^self outer!

ifCurtailed: terminationBlock
	"Evaluate the receiver, and should it exit via a non-local (^) return, or by raising an 
	exception, evaluate the niladic block, <terminationBlock>.
	Answers the result of evaluating the receiver, unless curtailed and the terminationBlock 
	contains a non-local return of its own, in which case the result of that block 
	will be answered to its home context's sender, e.g.:

		[^1] ifCurtailed: [Sound beep]			answers 1, but also woofs
		[^1] ifCurtailed: [Sound beep. ^2]		woofs and answers 2.
		[1] ifCurtailed: [Sound beep. ^2]		answers 1

	Be warned: The system may become unstable if this method is modified. The receiver is
	not the original receiver of the message after #valueOnUnwind: has been sent, but is
	valid inside the unwind block. In addition, #valueOnUnwind: leaves additional objects
	on the stack.

	See also #ensure:"

	| answer |
	answer := self valueOnUnwind: [:retValue :retFrame |
		terminationBlock value.
		Processor returnValue: retValue toFrame: retFrame].
	^answer!

initialIP
	"Private - Answer the initial instruction pointer index into the home method, 
	used when the block receives a #value* message"

	^initialIP!

initialIP: anInteger
	"Private - Set the initial instruction pointer index into the receiver's
	compiled byte codes (used when the block receives a #value* message)."

	initialIP := anInteger!

method
	"Private - Answer the method in which the receiver's code resides."

	^method!

method: aCompiledMethod
	"Private - Set the method in which the receiver's code resides."

	method := aCompiledMethod!

millisecondsToRepeat: anInteger
	"Answer the time taken to execute the receiver anInteger times"

	^Time millisecondsToRun: [ anInteger timesRepeat: self ]!

newProcess
	"Answer a new, suspended, Process which will evaluate the receiver, and terminate 
	when and if the code in the receiver returns. The default 'name' given to the process 
	is the originating CompiledMethod - this is useful for debugging because it 
	allows one to locate the source (e.g. Process name getSource) to determine 
	what the process is doing."

	^(Process forContext: [ 
			self on: ProcessTermination do: [:e | e return].
			"^-return catches attempt to drop off bottom of process stack (which would cause a GPF)"
			^Processor activeProcess shutdown]
		priority: Processor activePriority)
		name: self method;
		yourself
!

newProcessWithArguments: anArray
	"Answer a new process which will evaluate the receiver with arguments
	from anArray, then terminate"

	^[self valueWithArguments: anArray] newProcess!

on: selector do: action
	"Try to evaluate the receiver, and should an exception occur which is matched
	by the <exceptionSelector>, selector, normally a class object which is a subclass 
	of Exception), evaluate the <monadicBlock>, action, passing it the exception 
	instance as its argument."

	^(ExceptionHandler on: selector do: action) try: self!

on: selector1 do: action1 on: selector2 do: action2
	"Try to evaluate the receiver, and should an exception occur which is matched
	by any of the <exceptionSelector>s, selector1..selector2, then evaluate the 
	corresponding <monadicBlock>, actionN, passing it the exception instance as 
	its argument."

	^self onDo: 
		(ExceptionHandlerSet new
			on: selector1 do: action1;
			on: selector2 do: action2)!

on: selector1 do: action1 on: selector2 do: action2 on: selector3 do: action3
	"Try to evaluate the receiver, and should an exception occur which is matched
	by any of the <exceptionSelector>s, selector1..selector3, then evaluate the 
	corresponding <monadicBlock>, actionN, passing it the exception instance as 
	its argument."

	^self onDo: 
		(ExceptionHandlerSet new
			on: selector1 do: action1;
			on: selector2 do: action2;
			on: selector3 do: action3)!

on: selector1 do: action1 
	on: selector2 do: action2 
	on: selector3 do: action3
	on: selector4 do: action4
	"Try to evaluate the receiver, and should an exception occur which is matched
	by any of the <exceptionSelector>s, selector1..selector4, then evaluate the 
	corresponding <monadicBlock>, actionN, passing it the exception instance as 
	its argument."

	^self onDo: 
		(ExceptionHandlerSet new
			on: selector1 do: action1;
			on: selector2 do: action2;
			on: selector3 do: action3;
			on: selector4 do: action4)
!

onDo: exceptionHandler
	"Try to evaluate the receiver under the protection of the <ExceptionHandler>
	argument, exceptionHandler."

	^exceptionHandler try: self!

outer
	"Private - Answer the receiver's outer context."

	^outer!

outer: anObject
	"Private - Set the outer context of the receiver to anObject (a MethodContext or nil).
	Answer the receiver."

	outer := anObject!

printOn: aStream
	"Append, to aStream, a String whose characters are a description of the receiver."

	(self method isKindOf: CompiledExpression)
		ifTrue: [ "We can print the source"
			aStream nextPutAll: self method getSource ]
		ifFalse: [ "We can just show where the source is"
			aStream nextPutAll: '[] in '; 	print: self method ]!

receiver
	"Private - Answer the 'self' of the block."

	^self home receiver!

repeat
	"Evaluate the receiver repeatedly, ending only if the block throws some exception or 
	explicity returns.
	Note that this message is now inlined by the Compiler for literal block receivers."

	[self value. true] whileTrue!

speed
	"Copyright (c) 1993 by Bruce Samuelson. Released to the public domain.

	Measure the time to evaluate the receiver minus the time to evaluate an empty 
	block. Return a string detailing raw execution speed of the code inside the block.
	
	slow		['hello' , ' world'] speed
	medium	[1+1+1+1+1+1+1+1+1+1+1] speed
	fast			[1+1] speed
	infinite		[] speed

	To get accurate measures of raw speed for fast operations such as integer
	addition, the block must have many internal iterations. For example, VisualWorks
	1.0 on an Amax 486/33, 256K cache, 16 MB ram, DOS 6.0, Windows 3.1 yields

	1	million adds per second with [1+1]		one addition inside the block
	4-5	million adds per second with [1+1+...+1]	10 additions inside the block
	7	million adds per second with [1+1+...+1]	100 additions inside the block.

	The inability to get an accurate measurement with one internal iteration may be
	partly caused by the timeBlock being copying rather than clean. Copying blocks
	evaluate more slowly than the equivalent clean ones."

	| minSeconds targetSeconds n emptyBlock timeBlock secondsSelf secondsEmpty t |
	minSeconds := 0.1.
	targetSeconds := 1.0.
	n := 1.
	emptyBlock := [].
	timeBlock := [:bl | (Time millisecondsToRun: [n timesRepeat: bl]) * 1.0e-3].
	MemoryManager current collectGarbage.
	[(t := timeBlock value: self) < minSeconds] whileTrue: [n := n * 2].
	n := (n * targetSeconds / t) rounded.
	MemoryManager current collectGarbage.
	secondsSelf := timeBlock value: self.
	MemoryManager current collectGarbage.
	secondsEmpty := timeBlock value: emptyBlock.
	^(t := secondsSelf - secondsEmpty) > (secondsSelf * 0.15)
		ifTrue: ['
' , n printString , ' iterations
' , t printString , ' seconds (time self - time empty block)
' , (t * 1.0e6 / n) printString , ' microseconds per iteration
' , (n / t) printString , ' iterations per second']
		ifFalse: ['The block is too fast to measure this way.']!

value
	"Answer the result of evaluating the receiver. Fail if the receiver is not 
	a niladic (zero argument) block.
	
	This method is present to handle failures and as a target for #performs, as 
	invocations of #value are performed directly for zero argument blocks.

	Primitive failure reasons:
		0 -	the receiver does not expect 0 arguments."

	<primitive: 81>
	^self valueFailed: 0!

value: arg1
	"Answer the result of evaluating the receiver. Fail if the receiver is not 
	a monadic (one argument) block.
	
	This method is present to handle failures and as a target for #performs, as 
	invocations of #value are performed directly for one argument blocks.

	Primitive failure reasons:
		0 -	the receiver does not expect 1 argument."

	<primitive: 81>
	^self valueFailed: 1!

value: arg1 value: arg2
	"Answer the result of evaluating the receiver. Fail if the receiver is not 
	a dyadic (two argument) block.

	Primitive failure reasons:
		0 -	the receiver does not expect 2 arguments."

	<primitive: 81>
	^self valueFailed: 2!

value: arg1 value: arg2 value: arg3
	"Answer the result of evaluating the receiver. Fail if the receiver is not 
	a three argument block.

	Primitive failure reasons:
		0 -	the receiver does not expect 3 arguments."

	<primitive: 81>
	^self valueFailed: 3!

valueFailed: anInteger
	"Private - A primitive value call failed, generate an appropriate error"

	^self argumentCount == anInteger
		ifTrue: [super primitiveFailed]
		ifFalse: [self error: 'Block expects ', self argumentCount printString, ' argument(s)']!

valueOnUnwind: aTwoArgBlock
	"Private - Evaluate the receiver, and should it exit by a non-local return to its home method's sender
	(i.e. via a ^-return) then evaluate aTwoArgBlock, passing it the receiver's home context's sender (i.e.
	the return destination), and the return value. aTwoArgBlock must include an explicit return of its
	own, either via ProcessorScheduler>>return: to return the argument value to the argument context, or 
	via a further ^-return, which will override the result and return destinations.

	Be warned: The primitive marks the caller's stack frame as having an unwind block, by changing its 
	receiver to be a special cookie. The VM looks for this object when performing ^-returns from 
	blocks, and, on detecting such, locates and evaluates the unwind block. The current implementation of 
	this unwind mechanism leaves its unwind block argument on the stack before evaluating the receiver in 
	the normal way. It is recommended that user code should not send this message directly, as it is 
	possible that the VM may become unstable if this is not done correctly. User code should use one of 
	the two methods, #ensure: and #ifCurtailed: to record unwind blocks that are performed, 
	respectively, regardless of whether the block evalutes and returns normally or is curtailed by a non-
	local return or an exception, or only if curtailed.

	Primitive failure reasons:
		0 -	the receiver does not expect 0 arguments."

	<primitive: 79>
	^self primitiveFailed!

valueWithArguments: argumentArray
	"Answer the result of evaluating the receiver with arguments from
	the <Array>, argumentArray. Fail if the argument is not an <Array>, 
	or if the receiver does not take the same number of arguments as are 
	present in the <Array>.

	A block can have a maximum of 256 arguments. Most Smalltalks have lower limits,
	and for complete portability it is unwise to use more than two. Stack overflow
	is also a possibility when the arguments are transferred to the stack.

	Primitive failure reasons:
		0 -	the receiver does not expect the number of arguments in argumentArray.
		1 -	argumentArray is not an <Array>."

	<primitive: 82>
	^self valueFailed: argumentArray size!

valueWithArgumentsAt: anAddress descriptor: anExternalFunctionDescriptor 
	"Private - Answer the result of evaluating the receiver with arguments
	instantiated from memory at anAddress using the the external argument 
	types from, anExternalFunctionDescriptor.
	Implementation Note: For performance and consistency (with outbound external calls) 
	reasons, we use a VM supplied primitive to instantiate and push the arguments and 
	perform the message, but there is no reason that this cannot be done in Smalltalk 
	if different argument conversions are required. The standard conversions are the 
	same as those performed for return types by the external call primitive (see
	the ExternalLibrary class).

	Primitive Failure reasons:
		0	-	anExternalFunctionDescriptor is not a byte object.
		1	-	anAddress is not a valid address object (SmallInteger/byte indirection)
		2	-	the number of arguments whose types are specified by anExternalFunctionDescriptor
				does not match the number of arguments expected by the receiver."

	<primitive: 117>
	^self valueFailed: anExternalFunctionDescriptor argumentCount!

whileFalse
	"Repeatedly evaluate the receiver until it evaluates to true.
	Where the receiver is a zero argument literal block, this message
	is inlined by the compiler. This implementation is included for other cases 
	(though these are likely to be errors) and so that it may be #perform:'d. 
	The implementation appears to be recursive, but is itself inlined by the Compiler."

	^[self value] whileFalse!

whileFalse: iterationBlock
	"Repeatedly evaluate the receiver until it evaluates to true, and
	on each iteration evaluate the <niladicValuable> argument, iterationBlock. 
	Answer nil.
	N.B. Where the receiver and argument are zero argument literal blocks, the
	message is inlined by the Compiler (it generates a loop). This method is 
	included for other types of niladic valuable (though these are likely to 
	be errors) and so that it may be #perform:'d. The implementation appears 
	to be recursive, but is itself inlined by the Compiler."

	^[self value] whileFalse: [iterationBlock value]!

whileTrue
	"Repeatedly evaluate the receiver until it evaluates to false.
	Where the receiver is a zero argument literal block, this message
	is inlined by the compiler, this method is included for other cases 
	(though these are likely to be errors) and so that it may be #perform:'d. 
	The implementation appears to be recursive, but is itself inlined by the Compiler."

	^[self value] whileTrue!

whileTrue: iterationBlock
	"Repeatedly evaluate the receiver until it evaluates to false, and
	on each iteration evaluate the <niladicValuable> argument, iterationBlock. 
	Answer nil.
	N.B. Where the receiver and argument are zero argument literal blocks, the
	message is inlined by the Compiler (it generates a loop). This method is 
	included for other types of niladic valuabel (though these are likely to 
	be errors) and so that it may be #perform:'d. The implementation appears 
	to be recursive, but is itself inlined by the Compiler."

	^[self value] whileTrue: [iterationBlock value]! !

!BlockClosure categoriesFor: #argumentCount!accessing!ANSI protocols-valuable! !
!BlockClosure categoriesFor: #atPriority:!evaluating! !
!BlockClosure categoriesFor: #critical!scheduling! !
!BlockClosure categoriesFor: #ensure:!ANSI protocols-niladicBlock!unwinding! !
!BlockClosure categoriesFor: #fork!scheduling! !
!BlockClosure categoriesFor: #forkAt:!scheduling! !
!BlockClosure categoriesFor: #frameClass!constants! !
!BlockClosure categoriesFor: #home!accessing! !
!BlockClosure categoriesFor: #ifCurtailed:!ANSI protocols-niladicBlock!unwinding! !
!BlockClosure categoriesFor: #initialIP!accessing! !
!BlockClosure categoriesFor: #initialIP:!accessing! !
!BlockClosure categoriesFor: #method!accessing! !
!BlockClosure categoriesFor: #method:!accessing! !
!BlockClosure categoriesFor: #millisecondsToRepeat:!development! !
!BlockClosure categoriesFor: #newProcess!scheduling! !
!BlockClosure categoriesFor: #newProcessWithArguments:!scheduling! !
!BlockClosure categoriesFor: #on:do:!ANSI protocols-niladicBlock!exceptions! !
!BlockClosure categoriesFor: #on:do:on:do:!exceptions! !
!BlockClosure categoriesFor: #on:do:on:do:on:do:!exceptions! !
!BlockClosure categoriesFor: #on:do:on:do:on:do:on:do:!exceptions! !
!BlockClosure categoriesFor: #onDo:!exceptions! !
!BlockClosure categoriesFor: #outer!accessing! !
!BlockClosure categoriesFor: #outer:!accessing! !
!BlockClosure categoriesFor: #printOn:!printing! !
!BlockClosure categoriesFor: #receiver!accessing! !
!BlockClosure categoriesFor: #repeat!evaluating! !
!BlockClosure categoriesFor: #speed!public domain! !
!BlockClosure categoriesFor: #value!ANSI protocols-niladicValuable!evaluating! !
!BlockClosure categoriesFor: #value:!ANSI protocols-monadicValuable!evaluating! !
!BlockClosure categoriesFor: #value:value:!ANSI protocols-dyadicValuable!evaluating! !
!BlockClosure categoriesFor: #value:value:value:!evaluating! !
!BlockClosure categoriesFor: #valueFailed:!exceptions! !
!BlockClosure categoriesFor: #valueOnUnwind:!unwinding! !
!BlockClosure categoriesFor: #valueWithArguments:!ANSI protocols-valuable!evaluating! !
!BlockClosure categoriesFor: #valueWithArgumentsAt:descriptor:!evaluating! !
!BlockClosure categoriesFor: #whileFalse!ANSI protocols-niladicValuable!evaluating! !
!BlockClosure categoriesFor: #whileFalse:!ANSI protocols-niladicValuable!evaluating! !
!BlockClosure categoriesFor: #whileTrue!ANSI protocols-niladicValuable!evaluating! !
!BlockClosure categoriesFor: #whileTrue:!ANSI protocols-niladicValuable!evaluating! !

